<?php
namespace metagen_go;

use Exception;
use Twig\{TwigTest, TwigFilter, TwigFunction};

function codegen(?string $cache_dir, \mtgMetaInfo $meta, array $options = []) : string
{
  $twig = get_twig();

  if(!empty($cache_dir))
    $twig->setCache($cache_dir);

  $twig->addGlobal('meta', $meta);

  $options['meta'] = $meta;
  if(isset($options['package']))
    $options['package'] = 'autogen';

  return $twig->render('codegen_bundle.twig', $options);
}

function get_twig(array $inc_path = []): \Twig\Environment
{
  array_unshift($inc_path, __DIR__ . "/../tpl/");
  $loader = new \Twig\Loader\FilesystemLoader($inc_path);

  $twig = new \Twig\Environment($loader, [
    'debug' => true,
    'autoescape' => false,
    'strict_variables' => true,
  ]);
  $twig->addExtension(new \Twig\Extension\DebugExtension());

  add_twig_tests($twig);
  add_twig_filters($twig);
  add_twig_functions($twig);

  return $twig;
}

function supported_tokens()
{
  return [
    'table',
    'table_pkey',
    'table_json_kv',
    'db_skip',

    'data_root',
    'bitfields',
    'diffable',
    'diff_removed',

    'alias',
    'default',
    'virtual',

    'go_json_quote',

    'statist',
    'statist_skip',
    'statist_alias',
  ];
}

function add_twig_tests(\Twig\Environment $twig)
{
  $twig->addTest(new TwigTest(
    'metaenum',
    fn(\mtgMetaInfoUnit $u): bool => $u->object instanceof \mtgMetaEnum,
  ));
  $twig->addTest(new TwigTest(
    'metastruct',
    fn(\mtgMetaInfoUnit $u): bool => $u->object instanceof \mtgMetaStruct,
  ));
  $twig->addTest(new TwigTest(
    'metarpc',
    fn(\mtgMetaInfoUnit $u): bool => $u->object instanceof \mtgMetaRPC,
  ));
  $twig->addTest(new TwigTest(
    'enum',
    fn(\mtgType $t): bool => $t instanceof \mtgMetaEnum,
  ));
  $twig->addTest(new TwigTest(
    'struct',
    fn(\mtgType $t): bool => $t instanceof \mtgMetaStruct,
  ));
  $twig->addTest(new TwigTest(
    'array',
    fn(\mtgType $t): bool => $t instanceof \mtgArrType,
  ));
  $twig->addTest(new TwigTest(
    'builtin',
    fn(\mtgType $t): bool => $t instanceof \mtgBuiltinType,
  ));
  $twig->addTest(new TwigTest(
    'numeric',
    fn(\mtgType $t): bool => $t instanceof \mtgBuiltinType && $t->isNumeric(),
  ));
  $twig->addTest(new TwigTest(
    'int',
    fn(\mtgType $t): bool => $t instanceof \mtgBuiltinType && $t->isInt(),
  ));
  $twig->addTest(new TwigTest(
    'uint',
    fn(\mtgType $t): bool => $t instanceof \mtgBuiltinType && $t->isUint(),
  ));
  $twig->addTest(new TwigTest(
    'bool',
    fn(\mtgType $t): bool => $t instanceof \mtgBuiltinType && $t->isBool(),
  ));
  $twig->addTest(new TwigTest(
    'float',
    fn(\mtgType $t): bool => $t instanceof \mtgBuiltinType && $t->isFloat(),
  ));
  $twig->addTest(new TwigTest(
    'double',
    fn(\mtgType $t): bool => $t instanceof \mtgBuiltinType && $t->isDouble(),
  ));
  $twig->addTest(new TwigTest(
    'string',
    fn(\mtgType $t): bool => $t instanceof \mtgBuiltinType && $t->isString(),
  ));
  $twig->addTest(new TwigTest(
    'blob',
    fn(\mtgType $t): bool => $t instanceof \mtgBuiltinType && $t->isBlob(),
  ));

  $twig->addTest(new TwigTest(
    'instanceof',
    fn($obj, $class): bool => (new \ReflectionClass($class))->isInstance($obj),
  ));
}

function add_twig_filters(\Twig\Environment $twig)
{
  $twig->addFilter(new TwigFilter(
    'go_type',
    fn(\mtgType $t, array $tokens = []): string => go_type($t, $tokens),
  ));
  $twig->addFilter(new TwigFilter(
    'ucfirst',
    fn(string $str): string => ucfirst($str),
  ));
  $twig->addFilter(new TwigFilter(
    'lcfirst',
    fn(string $str): string => lcfirst($str),
  ));
  $twig->addFilter(new TwigFilter(
    'camel',
    fn(string $str): string => snake2camel($str),
  ));
  $twig->addFilter(new TwigFilter(
    'quote',
    fn(string $str, string $char = '`'): string => "{$char}{$str}{$char}"
  ));
  $twig->addFilter(new TwigFilter(
    'sql_where',
    function(array $fields): string {
      $exprs = [];
      foreach ($fields as $field) {
        $fieldName = $field->getName();
        $exprs[] = "`{$fieldName}` = ?";
      }
      return implode(' AND ', $exprs);
    }
  ));
  $twig->addFilter(new TwigFilter(
    'alias', 
    fn(\mtgMetaField $field): string => get_field_alias_or_name($field)
  ));
  $twig->addFilter(new TwigFilter(
    'fname', 
    fn(\mtgMetaField $field): string => snake2camel($field->getName())
  ));
  $twig->addFilter(new TwigFilter(
    'varname', 
    fn(\mtgMetaField $field): string => lcfirst(snake2camel($field->getName()))
  ));  
  $twig->addFilter(new TwigFilter(
    'tname', 
    fn(\mtgMetaField $field): string => $field->getType() instanceof \mtgArrType ? $field->getType()->getValue()->getName() : $field->getType()->getName()
  ));
  $twig->addFilter(new TwigFilter(
    'builtin_type_suffix', 
    fn (\mtgBuiltinType $type): string => builtin_type_method_suffix($type),
  ));
  $twig->addFilter(new TwigFilter(
    'default_val', 
    fn(\mtgMetaField $field) => default_value($field)
  ));
  $twig->addFilter(new TwigFilter(
    'fields', 
    function (\mtgMetaField $field): array {
      $type = $field->getType();
      if($type instanceof \mtgArrType)
      {
        $type = $type->getValue();
      }
      return $type->getFields();
    }
  ));
  $twig->addFilter(new TwigFilter(
    'pk_fields', 
    function (\mtgMetaField $field): array {
      $type = $field->getType();
      if($type instanceof \mtgArrType)
      {
        $type = $type->getValue();
      }
      $pkey = table_pkey_fields($type);
      $fields = [];
      foreach ($pkey as $fieldName) {
        $fields[] = $type->getField($fieldName);
      }
      return $fields;
    }
  ));
  $twig->addFilter(new TwigFilter(
    'raw_fields',
    function(\mtgMetaStruct $struct): array {
      $json_fields = table_json_fields($struct);
      return array_filter(
        $struct->getFields(),
        fn(\mtgMetaField $field): bool => !in_array($field->getName(), $json_fields),
      );
    }
  ));
  $twig->addFilter(new TwigFilter(
    'raw_nonpk_fields',
    function(\mtgMetaStruct $struct): array {
      $pkey = table_pkey_fields($struct);
      $json_fields = table_json_fields($struct);
      return array_filter(
        $struct->getFields(),
        fn(\mtgMetaField $field): bool => !in_array($field->getName(), $pkey) && !in_array($field->getName(), $json_fields),
      );
    }
  ));
  $twig->addFilter(new TwigFilter(
    'json_fields',
    function(\mtgMetaStruct $struct): array {
      $json_fields = table_json_fields($struct);
      return array_map(
        fn(string $field): \mtgMetaField => $struct->getField($field),
        $json_fields,
      );
    }
  ));
  $twig->addFilter(new TwigFilter(
    'table_columns',
    function(\mtgMetaStruct $struct): array {
      $json_fields = table_json_fields($struct);

      $columns = array_reduce(
        $struct->getFields(),
        function(array $carry, \mtgMetaField $field) use ($json_fields): array {
          if(in_array($field->getName(), $json_fields))
          {
            return $carry;
          }
          $carry[] = $field->getName();
          return $carry;
        },
        []
      );

      if(count($json_fields) > 0)
      {
        $columns[] = 'kv';
      }

      return $columns;
    }
  ));
  $twig->addFilter(new TwigFilter(
    'table_nopk_columns',
    function(\mtgMetaStruct $struct): array {
      $pkey = table_pkey_fields($struct);
      $json_fields = table_json_fields($struct);
      
      $columns = array_reduce(
        $struct->getFields(),
        function(array $carry, \mtgMetaField $field) use ($pkey, $json_fields): array {
          if(in_array($field->getName(), $pkey) || in_array($field->getName(), $json_fields))
          {
            return $carry;
          }
          $carry[] = $field->getName();
          return $carry;
        },
        []
      );

      if(count($json_fields) > 0)
      {
        $columns[] = 'kv';
      }

      return $columns;
    }
  ));
  $twig->addFilter(new TwigFilter(
    'table_fields',
    function(\mtgMetaStruct $struct): array {
      $json_fields = table_json_fields($struct);

      $fields = array_reduce(
        $struct->getFields(),
        function(array $carry, \mtgMetaField $field) use ($json_fields): array {
          if(in_array($field->getName(), $json_fields))
          {
            return $carry;
          }
          $carry[] = snake2camel($field->getName());
          return $carry;
        },
        []
      );

      if(count($json_fields) > 0)
      {
        $fields[] = json_field_name($struct);
      }

      return $fields;
    }
  ));
  $twig->addFilter(new TwigFilter(
    'json_tag',
    function(\mtgMetaField $field): string {
      $tag = get_field_alias_or_name($field);
      $type = $field->getType();

      //NOTE: currently quoted by default
      $quote = !$field->hasToken("go_json_quote") || 
        filter_var($field->getToken("go_json_quote"), FILTER_VALIDATE_BOOLEAN) === true; 

      if(!$quote)
        return $tag;

      if($type instanceof \mtgMetaEnum || 
        ($type instanceof \mtgBuiltinType && !$type->isString() && !$type->isBlob())) {
        return $tag.',string';
      }
      return $tag;
    }
  ));
}

function get_field_alias_or_name(\mtgMetaField $field): string
{
  if ($field->hasToken('alias')) {
    return $field->getToken('alias');
  }
  return $field->getName();
}

function add_twig_functions(\Twig\Environment $twig)
{
  $twig->addFunction(new TwigFunction(
    'Error',
    function(string $e) {
      throw new Exception($e);
    }
  ));
  $twig->addFunction(new TwigFunction(
    'has_token',
    fn($o, string $token): bool => $o->hasToken($token),
  ));
  $twig->addFunction(new TwigFunction(
    'has_token_in_parent',
    fn(\mtgMetaStruct $s, string $token): bool => $s->hasTokenInParent($token),
  ));
  $twig->addFunction(new TwigFunction(
    'token',
    fn($o, string $name) => $o->getToken($name),
  ));
  $twig->addFunction(new TwigFunction(
    'token_or',
    fn($o, string $name, $v) => $o->hasToken($name) ? $o->getToken($name) : $v,
  ));
  $twig->addFunction(new TwigFunction(
    'get_all_fields',
    fn(\mtgMetaStruct $s): array => \mtg_get_all_fields($s),
  ));
  $twig->addFunction(new TwigFunction(
    'arr_fill', 
    fn($value, int $count): array => array_fill(0, $count, $value)
  ));
  $twig->addFunction(new TwigFunction(
    'table_pkey', 
    function(\mtgMetaStruct $struct): array {
      $pkey = table_pkey_fields($struct);
      return array_map(
        fn(string $field): \mtgMetaField => $struct->getField($field),
        $pkey,
      );
    }
  ));
  $twig->addFunction(new TwigFunction(
    'table_fields', 
    function(\mtgMetaStruct $struct): array {
      $pkey = table_pkey_fields($struct);
      return array_filter(
        $struct->getFields(),
        fn(\mtgMetaField $field): bool => !in_array($field->getName(), $pkey),
      );
    }
  ));
  $twig->addFunction(new TwigFunction(
    'meta_field', 
    function (string $str) use ($twig): \mtgMetaField {
      $meta = $twig->getGlobals()['meta'];
      return parse_meta_field($str, $meta);
    }
  ));
}

function snake2camel(string $str): string
{
  return str_replace('_', '', ucwords($str, '_'));
}

function go_type(\mtgType $type, array $tokens = []): string
{
  if($type instanceof \mtgMetaEnum)
  {
    return $type->getName();
  }
  else if($type instanceof \mtgMetaStruct)
  {
    if(array_key_exists('virtual', $tokens)) {
      return 'I'.$type->getName();
    }
    return $type->getName();
  }
  else if($type instanceof \mtgBuiltinType) 
  {
    if($type->isFloat()) 
      return 'float32';
    else if($type->isDouble())
      return 'float64';
    else if($type->isBlob())
      return '[]byte';
    else
      return $type->getName();
  } 
  else if($type instanceof \mtgArrType) 
  {
    $native = go_type($type->getValue(), $tokens);
    return "[]$native";
  }
  else
    throw new Exception("Unknown type '{$type->getName()}'");
}

function builtin_type_method_suffix(\mtgBuiltinType $type): string
{
  if($type->isBlob())
    return 'Bytes';
  
  return ucfirst(go_type($type));
}

function default_value(\mtgMetaField $field)
{
  $type = $field->getType();
  $tokens = $field->getTokens();
  $default = $field->getToken('default');
  $is_virtual = $field->hasToken('virtual');

  if($type instanceof \mtgBuiltinType) 
  {
    if($type->isNumeric())
      return $default ?? 0;
    else if($type->isString())
      return $default ?? '""';
    else if($type->isBool())
      return $default ?? 'false';
    else if($type->isBlob())
    {
      if($default == 'null')
        $default = 'nil';
      
      return $default ?? 'nil';
    } 
    else
      throw new Exception("Unknown type '$type'");
    
  } 
  else if($type instanceof \mtgMetaEnum)
    return $default ? go_type($type, $tokens)."_".trim($tokens['default'], '"') : 0;
  else if($type instanceof \mtgMetaStruct)
  {
    if(!empty($default) && $default != 'null')
    {
      throw new Exception("not supported default value for struct '{$field->getName()}' with type '$type'");
    }
    return ($is_virtual ? '&' : '').go_type($type).'{}';
  }
  else
    throw new Exception("Unknown type '$type'");
}

function parse_meta_field(string $str, \mtgMetaInfo $meta): \mtgMetaField
{
  list($name, $type) = explode('|', $str);
  return new \mtgMetaField($name, new \mtgTypeRef(new \mtgBuiltinType($type)));
}

function table_pkey_fields(\mtgMetaStruct $struct): array
{
  if(!$struct->hasToken('table_pkey'))
  {
    return [];
  }
  $pkey = $struct->getToken('table_pkey');
  if(empty($pkey))
  {
    throw new Exception('`table_pkey` token is empty');
  }
  return explode(',', $pkey);
}

function table_json_fields(\mtgMetaStruct $struct): array
{
  if(!$struct->hasToken('table_json_kv'))
  {
    return [];
  }
  $json_kv = $struct->getToken('table_json_kv');
  if(empty($json_kv))
  {
    throw new Exception('`table_json_kv` token is empty');
  }
  return explode(',', $struct->getToken('table_json_kv'));
}

function json_field_name(\mtgMetaStruct $struct): string
{
  return $struct->getName() . 'JsonValues';
}
